

package worb.protocols.http;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Observer;

/** Servlet API */
import javax.servlet.ServletOutputStream;

import java.io.*;

/**
* @version 0.1
*/
public class 
HttpOutputStream extends ServletOutputStream implements HttpDebug
{

    //static final boolean DEBUG = false;

    protected static final int MAX_RESPONSE_SIZE = Integer.MAX_VALUE;
 
    /** OutputStream on which we are gonna write */

    protected OutputStream out;

    /** Write buffer that we are gonna use */
    protected byte buf[];

    protected int count;

    protected int total;

    protected int limit;

    protected int length;

    /** Observer object */
    protected Observer obs;

    protected IOException except;

    protected boolean committed;

    private int countedLength;
    ByteArrayOutputStream bout;

    public 
    HttpOutputStream(int i)
    {
        buf = new byte[i];
        setContentLength(i);
        bout = new ByteArrayOutputStream();
    }

    public 
    HttpOutputStream()
    {
        this(20000);
        bout = new ByteArrayOutputStream();
    }

    public void 
    init(OutputStream outputStream) throws IOException
    {
        out = outputStream;
    }

    public OutputStream 
    getOutputStream() 
    {
        return out;
    }

    public PrintWriter 
    getWriter() 
    {
        return new PrintWriter(out);
    }

    /**
    * Get ready for the next HTTP response.
    */
    public void 
    next()
    {
        count = 0;
        total = 0;
        limit = -1;
        length = -1;
        obs = null;
        committed = false;
        countedLength = 0;
        bout = new ByteArrayOutputStream();
    }

    public void 
    finish() throws IOException
    {
        if (length == -1 && countedLength != 0)
            length = countedLength;
        flush();
        countedLength = 0;
        length = 0;
    }

    /** 
    * Clears the I/O related objects.
    */
    public void 
    resets()
    {
        out = null;
        obs = null;
        bout = new ByteArrayOutputStream();
    }

    public int 
    getTotal()
    {
        return total;
    }

    public void 
    setContentLength(int i)
    {
        if (i < 0)
            throw new IllegalArgumentException("invalid content length");
        length = i;
    }

    /**
    * @return the number of bytes so far written to the output stream.
    */
    public int 
    getContentLength()
    {
        return length;
    }

    /** 
    * Sets the observer of this object.
    */
    public void 
    setObserver(Observer observer)
    {
        obs = observer;
        limit = -1;
    }

    /**
    * Sets the IOException.
    */ 
    public void 
    setIOException(IOException e)
    {
        except = e;
        limit = -1;
    }

    public boolean 
    isCommitted()
    {
        return committed;
    }

    protected void 
    check(int i) throws IOException
    {
        /*******************************************************/
        /** There is an Exception raised before, then throw it */
        /*******************************************************/
        if (except != null) throw except;

        //if(DEBUG) System.out.println("except is NULL!");
        if (limit < 0)
        {
            if (total > 0) throw new InternalError();
            limit = MAX_RESPONSE_SIZE;
            if (except != null)
            {
                limit = -1;
                throw except;
            }
            //if(DEBUG) System.out.println(" check0: limit : "+limit); 
            //if(DEBUG) System.out.println(" check1: length : "+length); 
            if (length != -1 && length < MAX_RESPONSE_SIZE - total)
                limit = total + length;
            //if(DEBUG) System.out.println(" check1: limit : "+limit); 
        }
        //if (total + i <= limit)
        //limit = -1;
        //throw except = new IOException("write past end of stream");

        if (total + i > limit)
        { limit = -1;
          throw except = new IOException("write past end of stream");
        }// end of if

    }// end of check()

    public void 
    write(int i) throws IOException
    {
        if (total >= limit) check(1);
        if (count == buf.length) flushBytes();
        else if(DEBUG) System.out.println(" I did not write! ");
        buf[count++] = (byte)i;
        total++;
        countedLength++;
    }

    public void 
    write(byte ab[], int i, int j) throws IOException
    {
        if (j < 0)
            throw new IllegalArgumentException();
        //if(DEBUG) System.out.println(" limit = "+limit);
        //if(DEBUG) System.out.println(" total = "+total);
        //if(DEBUG) System.out.println(" j = "+j);
        if (total + j > limit) check(j);
        if (j >= buf.length)
        {
            flushBytes();
            total += j;
            writeOut(ab, i, j);
            return;
        }
        int k = buf.length - count;
        if (j > k)
            flushBytes();
        System.arraycopy(ab, i, buf, count, j);
        count += j;
        total += j;
        countedLength += j;
    }

    public void 
    flush() throws IOException
    {
        if(DEBUG) System.out.println(" I am in flush ");
        if (limit < 0) check(0);
        if(DEBUG) System.out.println(" I am before flushBytes ");
        flushBytes();
        if(DEBUG) System.out.println(" I am after flushBytes ");
    }

    protected void 
    flushBytes() throws IOException
    {
        //if(DEBUG) 
        System.out.print(" I am in flushBytes count is "+count);
        System.out.println(" contentlength "+getContentLength());
        if (!committed && obs != null)
        {
            if(DEBUG) System.out.println(" In flushBytes() of HttpOutputStream ");
            obs.update(null, this);
        }
        committed = true;
        if (count > 0)
        {
            //writeOut(buf, 0, count);
            writeOut(bout.toByteArray(), 0, bout.size());
            out.flush();
            count = 0;
        }
        else if(DEBUG) System.out.println(" I am in flushBytes count is "+count);
    }

    ////***** BEGINNING OF java.servlet.ServletOutputStream Interface *****/

    public void 
    print(String string) throws IOException
    {
        /*
        if(DEBUG) System.out.println(" limit : "+limit+" string : "+string);
        int i = string.length();
        if (total + i > limit) check(i);
        int j = 0;
        if(DEBUG) System.out.println(" limit : "+limit);
        while (i > 0)
        {
            int k = buf.length - count;
            if (k == 0)
            {
                flushBytes();
                k = buf.length - count;
            }
            if (k > i)
                k = i;
            string.getBytes(j, j + k, buf, count);
            count += k;
            total += k;
            j += k;
            i -= k;
            countedLength += k;
        }
        */

        count += string.length();
        countedLength = string.length();
        bout.write(string.getBytes());
        setContentLength(bout.size());
    }

    public void
    println(String str)  throws IOException
    {
       print(str);println();
    }

    public void 
    print(HttpDate httpDate) throws IOException
    {
        int i = HttpDate.DATELEN;
        if (total + i > limit) check(i);
        int j = buf.length - count;
        if (j < i)
        {
            flushBytes();
            j = buf.length - count;
        }
        i = httpDate.getBytes(buf, count, j);
        count += i;
        total += i;
    }

    public void
    print(boolean val) throws IOException
    { print((new Boolean(val)).toString()); 
    }// end of print()

    public void
    println(boolean val) throws IOException
    { print(val);println(); 
    }// end of print()

    public void
    print(char val) throws IOException
    { print( (new Character(val)).toString()); 
    }// end of print()

    public void
    println(char val) throws IOException
    { print(val);println(); 
    }// end of print()

    public void
    print(double val) throws IOException
    {  print(Double.toString(val)); 
    }// end of print()

    public void
    println(double val) throws IOException
    { print(val);println(); 
    }// end of print()

    public void
    print(float val) throws IOException
    { print(Float.toString(val)); 
    }// end of print()

    public void
    println(float val) throws IOException
    { print(val);println(); 
    }// end of print()

    public void
    print(long val) throws IOException
    { print(val);println(); 
    }// end of print()

    public void
    println(long val) throws IOException
    {
       print(Long.toString(val)); 
    }// end of print()

    public void
    print(int val) throws IOException
    {
       print(Integer.toString(val)); 
    }// end of print()

    public void
    println(int val) throws IOException
    { print(val);println(); 
    }// end of print()

    public void
    println() throws IOException
    {
      print("\r\n"); 
    }// end of print()

    ////***** END OF java.servlet.ServletOutputStream Interface *****/

    public void 
    close() throws IOException
    {
        if(DEBUG) System.out.println(" I am in close ");
        finish();
    }

    /**
    * Writes the Header.
    */
    public void 
    setHeader(byte ab[]) throws IOException
    {
        String line = new String(ab);
        if(DEBUG) System.out.println(" header :"+line);
        writeOut(ab, 0, ab.length);
    }

    /**
    * Writes the Header.
    */
    public void 
    setHeader(byte ab[], int i, int j) throws IOException
    {
        String line = new String(ab);
        if(DEBUG) System.out.println(" header :"+line);
        writeOut(ab, i, j);
    }

    /**
    * Writes the given byte array to the OutputStream.
    */
    protected void 
    writeOut(byte ab[], int i, int j) throws IOException
    {
        String line = new String(ab);
        if(DEBUG) System.out.println(" line :"+line);
        out.write(ab, i, j);
    }

}// end of HttpOutputStream



